@@ -13,33 +13,82 @@
 #  include <google/protobuf/descriptor.pb.h>
 #endif
 
+#include <array>
 #include <cassert>
-#include <algorithm>
+#include <string_view>
 
 using namespace ::google::protobuf;
 using namespace ::google::protobuf::io;
 using namespace ::qtprotoccommon;
 
+namespace {
+/*
+    Returns the default initializer string for the corresponding field type.
+*/
+std::string_view getInitializerByFieldType(FieldDescriptor::Type type)
+{
+    static constexpr std::string_view FloatingPointInitializer = "0.0";
+    static constexpr std::string_view IntegerInitializer = "0";
+    static constexpr std::string_view BoolInitializer = "false";
+    switch (type) {
+    case FieldDescriptor::TYPE_DOUBLE:
+    case FieldDescriptor::TYPE_FLOAT:
+        return FloatingPointInitializer;
+    case FieldDescriptor::TYPE_FIXED32:
+    case FieldDescriptor::TYPE_FIXED64:
+    case FieldDescriptor::TYPE_INT32:
+    case FieldDescriptor::TYPE_INT64:
+    case FieldDescriptor::TYPE_SINT32:
+    case FieldDescriptor::TYPE_SINT64:
+    case FieldDescriptor::TYPE_UINT32:
+    case FieldDescriptor::TYPE_UINT64:
+        return IntegerInitializer;
+    case FieldDescriptor::TYPE_BOOL:
+        return BoolInitializer;
+    default:
+        break;
+    }
+    return {};
+}
+} // namespace
+
+std::set<std::string> common::m_extraNamespacedFiles;
+
 /*
     Constructs a C++ namespace from the full protobuf descriptor name. E.g. for
     the message descriptor "test.protobuf.MessageType" the function
     returns "test::protobuf", if the separator is "::".
 */
 std::string common::getFullNamespace(std::string_view fullDescriptorName,
-                                     std::string_view separator)
+                                     std::string_view separator,
+                                     bool extraScope)
 {
-    std::string output = Options::instance().extraNamespace();
+    std::string output = (extraScope) ? Options::instance().extraNamespace() : "";
     std::string::size_type nameIndex = fullDescriptorName.rfind('.');
     if (nameIndex == std::string::npos)
         return output;
     std::string namespacesStr =
             utils::replace(fullDescriptorName.substr(0, nameIndex), ".", separator);
+    if (namespacesStr == "QtCore" || namespacesStr == "QtGui")
+        namespacesStr = "QtProtobufPrivate" + std::string(separator) + namespacesStr;
     if (!output.empty() && !namespacesStr.empty())
         output += separator;
     output += namespacesStr;
     return output;
 }
 
+std::string common::buildExportMacro(bool addTrailingSpace)
+{
+    static const std::string &macroOption = Options::instance().exportMacro();
+    if (macroOption.empty())
+        return macroOption;
+
+    static std::string macro = "QPB_" + macroOption + "_EXPORT";
+    std::string resultingMacro = macro;
+    if (addTrailingSpace)
+        resultingMacro += ' ';
+    return resultingMacro;
+}
 /*
     Constructs a C++ namespace for wrapping nested message types.
     E.g. for the message descriptor with name "test.protobuf.MessageType.NestedMessageType" the
@@ -49,13 +98,14 @@ std::string common::getFullNamespace(std::string_view fullDescriptorName,
 std::string common::getNestedNamespace(const Descriptor *type, std::string_view separator)
 {
     assert(type != nullptr);
-    std::string namespaces = type->file()->package();
+    std::string namespaces{ type->file()->package() };
 
     std::string nestingNamespaces;
     const Descriptor *containingType = type->containing_type();
     while (containingType) {
         nestingNamespaces.insert(0,
-                                 std::string(separator) + containingType->name()
+                                 std::string(separator)
+                                         + utils::capitalizeAsciiName(containingType->name())
                                          + CommonTemplates::QtProtobufNestedNamespace());
         containingType = containingType->containing_type();
     }
@@ -97,32 +147,64 @@ std::map<std::string, std::string> common::getNestedScopeNamespace(const std::st
 
 TypeMap common::produceQtTypeMap(const Descriptor *type, const Descriptor *scope)
 {
-    std::string namespaces = getFullNamespace(type, "::");
     std::string scopeNamespaces = getScopeNamespace(type, scope);
     std::string qmlPackage = getFullNamespace(type, ".");
 
-    std::string name = type->name();
+    std::string name{ type->name() };
     std::string fullName = name;
     std::string scopeName = name;
 
-    std::string listName = std::string("QList<") + CommonTemplates::RepeatedSuffix() + ">";
-    std::string fullListName = listName;
-    std::string scopeListName = listName;
+    std::string listName = "QList<" + name + ">";
 
     return { { "type", name },
              { "full_type", fullName },
              { "scope_type", scopeName },
              { "list_type", listName },
-             { "full_list_type", fullListName },
-             { "scope_list_type", scopeListName },
+             { "full_list_type", listName },
+             { "scope_list_type", listName },
              { "scope_namespaces", scopeNamespaces },
              { "qml_package", qmlPackage },
              { "property_type", fullName },
-             { "property_list_type", fullListName },
+             { "property_list_type", listName },
              { "getter_type", scopeName },
              { "setter_type", scopeName } };
 }
 
+TypeMap common::produceOverriddenTypeMap(const Descriptor *type, const Descriptor *scope)
+{
+    std::string namespaces = getFullNamespace(type, "::");
+    std::string qmlPackage = getFullNamespace(type, ".");
+
+    std::string name{ type->name() };
+    std::string listName;
+    if (type->full_name() == "google.protobuf.Any") {
+        namespaces = "QtProtobuf";
+        name = "QtProtobuf::Any";
+        listName = std::string("QList<QtProtobuf::Any>");
+        qmlPackage = "QtProtobuf";
+    } else {
+        listName = std::string("QList<") + name + ">";
+    }
+    const std::string scopeNamespaces =
+            getScopeNamespace(namespaces, getFullNamespace(scope, "::"));
+    const std::string fullName = name;
+    const std::string scopeName = fullName;
+
+    return { { "type", name },
+             { "full_type", fullName },
+             { "scope_type", scopeName },
+             { "list_type", listName },
+             { "full_list_type", listName },
+             { "scope_list_type", listName },
+             { "scope_namespaces", scopeNamespaces },
+             { "qml_package", qmlPackage },
+             { "property_type", fullName },
+             { "property_list_type", listName },
+             { "getter_type", scopeName },
+             { "setter_type", scopeName },
+             { "initializer", "{}" } };
+}
+
 TypeMap common::produceMessageTypeMap(const Descriptor *type, const Descriptor *scope)
 {
     std::string namespaces = getFullNamespace(type, "::");
@@ -136,30 +218,30 @@ TypeMap common::produceMessageTypeMap(const Descriptor *type, const Descriptor *
     std::string fullName = namespaces.empty() ? name : (namespaces + "::" + name);
     std::string scopeName = scopeNamespaces.empty() ? name : (scopeNamespaces + "::" + name);
 
-    std::string listName = name + CommonTemplates::RepeatedSuffix();
-    std::string fullListName = namespaces.empty() ? listName : (namespaces + "::" + listName);
-    std::string scopeListName =
-            scopeNamespaces.empty() ? listName : (scopeNamespaces + "::" + listName);
+    std::string listName = "QList<" + name + ">";
+    std::string fullListName = namespaces.empty() ? listName : "QList<" + fullName + ">";
+    std::string scopeListName = scopeNamespaces.empty() ? listName : "QList<" + scopeName + ">";
 
-    std::string exportMacro = Options::instance().exportMacro();
-    exportMacro = common::buildExportMacro(exportMacro);
+    std::string exportMacro = common::buildExportMacro();
 
-    return {
-        { "classname", name },
-        { "type", name },
-        { "full_type", fullName },
-        { "scope_type", scopeName },
-        { "list_type", listName },
-        { "full_list_type", fullListName },
-        { "scope_list_type", scopeListName },
-        { "scope_namespaces", scopeNamespaces },
-        { "qml_package", qmlPackage },
-        { "property_type", fullName },
-        { "property_list_type", fullListName },
-        { "getter_type", scopeName },
-        { "setter_type", scopeName },
-        { "export_macro", exportMacro },
-    };
+    const std::string initializer = "nullptr";
+    return { { "classname", name },
+             { "dataclassname", name + CommonTemplates::DataClassName() },
+             { "classname_low_case", utils::deCapitalizeAsciiName(type->name()) },
+             { "type", name },
+             { "full_type", fullName },
+             { "scope_type", scopeName },
+             { "list_type", listName },
+             { "full_list_type", fullListName },
+             { "scope_list_type", scopeListName },
+             { "scope_namespaces", scopeNamespaces },
+             { "qml_package", qmlPackage },
+             { "property_type", fullName },
+             { "property_list_type", fullListName },
+             { "getter_type", scopeName },
+             { "setter_type", scopeName },
+             { "export_macro", exportMacro },
+             { "initializer", initializer } };
 }
 
 TypeMap common::produceEnumTypeMap(const EnumDescriptor *type, const Descriptor *scope)
@@ -168,7 +250,7 @@ TypeMap common::produceEnumTypeMap(const EnumDescriptor *type, const Descriptor
     std::string namespaces = getFullNamespace(type, "::");
 
     std::string name = utils::capitalizeAsciiName(type->name());
-    // qml package should consist only from proto spackage
+    // qml package should consist only from proto package
     std::string qmlPackage = getFullNamespace(type, ".");
     if (qmlPackage.empty())
         qmlPackage = "QtProtobuf";
@@ -185,34 +267,33 @@ TypeMap common::produceEnumTypeMap(const EnumDescriptor *type, const Descriptor
     std::string fullName = namespaces.empty() ? name : (namespaces + "::" + name);
     std::string scopeName = scopeNamespaces.empty() ? name : (scopeNamespaces + "::" + name);
 
-    std::string listName = name + CommonTemplates::RepeatedSuffix();
-    std::string fullListName = namespaces.empty() ? listName : (namespaces + "::" + listName);
-    std::string scopeListName =
-            scopeNamespaces.empty() ? listName : (scopeNamespaces + "::" + listName);
+    std::string listName = "QList<" + name + '>';
+    std::string fullListName = namespaces.empty() ? listName : "QList<" + fullName + '>';
+    std::string scopeListName = scopeNamespaces.empty() ? listName : "QList<" + scopeName + '>';
 
     // Note: For local enum classes it's impossible to use class name space in Q_PROPERTY
     // declaration. So please avoid addition of namespaces in line bellow
     std::string propertyType = visibility == LOCAL_ENUM ? name : fullName;
-    std::string exportMacro = Options::instance().exportMacro();
-    exportMacro = common::buildExportMacro(exportMacro);
+    std::string exportMacro =  common::buildExportMacro();
 
-    return {
-        { "classname", enumGadget },
-        { "type", name },
-        { "full_type", fullName },
-        { "scope_type", scopeName },
-        { "list_type", listName },
-        { "full_list_type", fullListName },
-        { "scope_list_type", scopeListName },
-        { "scope_namespaces", scopeNamespaces },
-        { "qml_package", qmlPackage },
-        { "property_type", propertyType },
-        { "property_list_type", fullListName },
-        { "getter_type", scopeName },
-        { "setter_type", scopeName },
-        { "enum_gadget", enumGadget },
-        { "export_macro", exportMacro },
-    };
+    std::string initializer = scopeName + "::" + common::qualifiedCppName(type->value(0)->name());
+    return { { "classname", name },
+             { "classname_low_case", utils::deCapitalizeAsciiName(name) },
+             { "type", name },
+             { "full_type", fullName },
+             { "scope_type", scopeName },
+             { "list_type", listName },
+             { "full_list_type", fullListName },
+             { "scope_list_type", scopeListName },
+             { "scope_namespaces", scopeNamespaces },
+             { "qml_package", qmlPackage },
+             { "property_type", propertyType },
+             { "property_list_type", fullListName },
+             { "getter_type", scopeName },
+             { "setter_type", scopeName },
+             { "enum_gadget", enumGadget },
+             { "export_macro", exportMacro },
+             { "initializer", initializer } };
 }
 
 TypeMap common::produceSimpleTypeMap(FieldDescriptor::Type type)
@@ -240,21 +321,7 @@ TypeMap common::produceSimpleTypeMap(FieldDescriptor::Type type)
     if (type != FieldDescriptor::TYPE_STRING && type != FieldDescriptor::TYPE_BYTES)
         fullListName = CommonTemplates::QtProtobufNamespace() + "::"s + listName;
     std::string scopeListName = fullListName;
-
-    std::string qmlAliasType;
-    switch (type) {
-    case FieldDescriptor::TYPE_INT32:
-    case FieldDescriptor::TYPE_SFIXED32:
-        qmlAliasType = "int";
-        break;
-    case FieldDescriptor::TYPE_FIXED32:
-        qmlAliasType = "unsigned int";
-        break;
-    default:
-        qmlAliasType = fullName;
-        break;
-    }
-
+    const std::string initializer(getInitializerByFieldType(type));
     return { { "type", name },
              { "full_type", fullName },
              { "scope_type", fullName },
@@ -264,39 +331,71 @@ TypeMap common::produceSimpleTypeMap(FieldDescriptor::Type type)
              { "scope_namespaces", namespaces },
              { "qml_package", qmlPackage },
              { "property_type", fullName },
-             { "qml_alias_type", qmlAliasType },
              { "property_list_type", fullListName },
              { "getter_type", fullName },
-             { "setter_type", fullName } };
+             { "setter_type", fullName },
+             { "initializer", initializer } };
 }
 
 MethodMap common::produceMethodMap(const MethodDescriptor *method, const std::string &scope)
 {
-    std::string inputTypeName = method->input_type()->full_name();
-    std::string outputTypeName = method->output_type()->full_name();
-    std::string methodName = method->name();
-    std::string methodNameUpper = method->name();
-    methodNameUpper[0] = static_cast<char>(::toupper(methodNameUpper[0]));
-    inputTypeName = utils::replace(inputTypeName, ".", "::");
-    outputTypeName = utils::replace(outputTypeName, ".", "::");
-    return { { "classname", scope },          { "return_type", outputTypeName },
-             { "method_name", methodName },   { "method_name_upper", methodNameUpper },
-             { "param_type", inputTypeName }, { "param_name", "arg" },
-             { "return_name", "ret" } };
+    std::string inputTypeName{ utils::replace(method->input_type()->full_name(), ".", "::") };
+    std::string outputTypeName{ utils::replace(method->output_type()->full_name(), ".", "::")  };
+    std::string methodName{ method->name() };
+    std::string methodNameUpper{ method->name() };
+    methodNameUpper[0] = static_cast<char>(utils::toAsciiUpper(methodNameUpper[0]));
+
+    std::string senderName = std::move(methodNameUpper) + "Sender";
+
+    std::string serviceName{ method->service()->name() };
+
+    //Make sure that we don't clash the same stream names from different services
+    std::string senderQmlName = serviceName;
+    senderQmlName += senderName;
+
+    std::string streamType;
+    if (method->client_streaming() && method->server_streaming()) {
+        streamType = "Bidi";
+    } else if (method->server_streaming()) {
+        streamType = "Server";
+    } else if (method->client_streaming()) {
+        streamType = "Client";
+    }
+
+    std::string streamTypeLower = utils::deCapitalizeAsciiName(streamType);
+
+    static const std::string exportMacro = common::buildExportMacro();
+
+    return {
+        {"classname",           scope                              },
+        { "sender_class_name",  senderName                         },
+        { "sender_qml_name",    senderQmlName                      },
+        { "return_type",        outputTypeName                     },
+        { "classname_low_case", utils::deCapitalizeAsciiName(scope)},
+        { "method_name",        methodName                         },
+        { "param_type",         inputTypeName                      },
+        { "param_name",         "arg"                              },
+        { "stream_type",        streamType                         },
+        { "stream_type_lower",  streamTypeLower                    },
+        { "return_name",        "ret"                              },
+        { "export_macro",       exportMacro                        },
+        { "service_name",       serviceName                        },
+    };
 }
 
 TypeMap common::produceServiceTypeMap(const ServiceDescriptor *service, const Descriptor *scope)
 {
-    const std::string name = "Service";
-    const std::string fullName = "Service";
-    const std::string scopeName = service->name();
-    const std::string exportMacro = common::buildExportMacro(Options::instance().exportMacro());
+    const std::string name{ "Service" };
+    const std::string fullName{ "Service" };
+    const std::string scopeName{ service->name() };
+    static const std::string exportMacro = common::buildExportMacro();
 
     const std::string namespaces = getFullNamespace(service, "::");
     const std::string scopeNamespaces = getScopeNamespace(namespaces,
                                                           getFullNamespace(scope, "::"));
 
     return { { "classname", name },
+             { "classname_low_case", utils::deCapitalizeAsciiName(name) },
              { "full_type", fullName },
              { "scope_type", scopeName },
              { "scope_namespaces", scopeNamespaces },
@@ -306,34 +405,65 @@ TypeMap common::produceServiceTypeMap(const ServiceDescriptor *service, const De
 
 TypeMap common::produceClientTypeMap(const ServiceDescriptor *service, const Descriptor *scope)
 {
-    const std::string name = "Client";
-    const std::string fullName = "Client";
-    const std::string scopeName = service->name();
-    const std::string exportMacro = common::buildExportMacro(Options::instance().exportMacro());
+    const std::string name{ "Client" };
+    const std::string fullName{ "Client" };
+    const std::string scopeName{ service->name() };
+    static const std::string exportMacro = common::buildExportMacro();
+
+    const std::string namespaces = getFullNamespace(service, "::");
+    const std::string scopeNamespaces = getScopeNamespace(namespaces,
+                                                          getFullNamespace(scope, "::"));
+
+    const std::string serviceName{ service->full_name() };
+    return { { "classname", name },
+             { "classname_low_case", utils::deCapitalizeAsciiName(name) },
+             { "full_type", fullName },
+             { "scope_type", scopeName },
+             { "scope_namespaces", scopeNamespaces },
+             { "parent_class", "QGrpcClientBase" },
+             { "service_name", serviceName },
+             { "export_macro", exportMacro } };
+}
+
+TypeMap common::produceQmlClientTypeMap(const ServiceDescriptor *service, const Descriptor *scope)
+{
+    const std::string name{ "QmlClient" };
+    const std::string fullName{ "QmlClient" };
+    const std::string serviceName{ service->name() };
+    static const std::string exportMacro = common::buildExportMacro();
 
     const std::string namespaces = getFullNamespace(service, "::");
     const std::string scopeNamespaces = getScopeNamespace(namespaces,
                                                           getFullNamespace(scope, "::"));
 
     return { { "classname", name },
+             { "classname_low_case", utils::deCapitalizeAsciiName(name) },
              { "full_type", fullName },
-             { "scope_type", scopeName },
+             { "scope_type", serviceName },
+             { "service_name", serviceName },
              { "scope_namespaces", scopeNamespaces },
-             { "parent_class", "QAbstractGrpcClient" },
+             { "parent_class", "Client" },
              { "export_macro", exportMacro } };
 }
 
 bool common::isQtType(const FieldDescriptor *field)
 {
-    return utils::startsWith(field->message_type()->full_name(), "QtProtobuf.")
-            && field->file()->package() != "QtProtobuf"; // Used for qttypes library to avoid types
-                                                         // conversion inside library
+    const auto fullName = field->message_type()->full_name();
+    const auto package = field->file()->package();
+    return (utils::startsWith(fullName, "QtCore.") || utils::startsWith(fullName, "QtGui."))
+            && package != "QtCore" && package != "QtGui";
+}
+
+bool common::isOverridden(const FieldDescriptor *field)
+{
+    return field->type() == FieldDescriptor::TYPE_MESSAGE
+            && field->message_type()->full_name() == "google.protobuf.Any";
 }
 
 bool common::isPureMessage(const FieldDescriptor *field)
 {
     return field->type() == FieldDescriptor::TYPE_MESSAGE && !field->is_map()
-            && !field->is_repeated() && !common::isQtType(field);
+            && !field->is_repeated() && !common::isQtType(field) && !common::isOverridden(field);
 }
 
 void common::iterateMessageFields(const Descriptor *message, const IterateMessageLogic &callback)
@@ -346,14 +476,42 @@ void common::iterateMessageFields(const Descriptor *message, const IterateMessag
     }
 }
 
+void common::iterateOneofFields(const Descriptor *message, const IterateOneofCallback &callback)
+{
+    int numFields =
+#ifdef HAVE_REAL_ONEOF_DECL
+        message->real_oneof_decl_count();
+#else
+        message->oneof_decl_count();
+#endif
+
+    for (int i = 0; i < numFields; ++i) {
+        const OneofDescriptor *field =
+#ifdef HAVE_REAL_ONEOF_DECL
+            message->real_oneof_decl(i);
+#else
+            message->oneof_decl(i);
+#endif
+#if defined(HAVE_PROTOBUF_SYNC_PIPER) && !defined(HAVE_REAL_ONEOF_DECL)
+        if (field->is_synthetic())
+            continue;
+#endif
+        auto propertyMap = common::producePropertyMap(field, message);
+        callback(field, propertyMap);
+    }
+}
+
 TypeMap common::produceTypeMap(const FieldDescriptor *field, const Descriptor *scope)
 {
     assert(field != nullptr);
 
     switch (field->type()) {
     case FieldDescriptor::TYPE_MESSAGE:
-        return isQtType(field) ? produceQtTypeMap(field->message_type(), nullptr)
-                               : produceMessageTypeMap(field->message_type(), scope);
+        if (isQtType(field))
+            return produceQtTypeMap(field->message_type(), nullptr);
+        if (isOverridden(field))
+            return produceOverriddenTypeMap(field->message_type(), nullptr);
+        return produceMessageTypeMap(field->message_type(), scope);
     case FieldDescriptor::TYPE_ENUM:
         return produceEnumTypeMap(field->enum_type(), scope);
     default:
@@ -363,9 +521,27 @@ TypeMap common::produceTypeMap(const FieldDescriptor *field, const Descriptor *s
     return produceSimpleTypeMap(field->type());
 }
 
+PropertyMap common::producePropertyMap(const OneofDescriptor *oneof, const Descriptor *scope)
+{
+    assert(oneof != nullptr);
+    assert(scope != nullptr);
+
+    PropertyMap propertyMap;
+    propertyMap["optional_property_name"] = qualifiedCppName(qualifiedQmlName(oneof->name()));
+    propertyMap["optional_property_name_cap"] = utils::capitalizeAsciiName(oneof->name());
+    auto scopeTypeMap = produceMessageTypeMap(scope, nullptr);
+    propertyMap["classname"] = scopeTypeMap["classname"];
+    propertyMap["dataclassname"] = propertyMap["classname"] + CommonTemplates::DataClassName();
+    propertyMap["type"] = propertyMap["optional_property_name_cap"] + "Fields";
+    propertyMap["export_macro"] = common::buildExportMacro();
+
+    return propertyMap;
+}
+
 PropertyMap common::producePropertyMap(const FieldDescriptor *field, const Descriptor *scope)
 {
     assert(field != nullptr);
+    assert(scope != nullptr);
 
     PropertyMap propertyMap = produceTypeMap(field, scope);
 
@@ -378,19 +554,32 @@ PropertyMap common::producePropertyMap(const FieldDescriptor *field, const Descr
         scriptable = "false";
     }
 
-    std::string propertyName = qualifiedName(utils::deCapitalizeAsciiName(field->camelcase_name()));
+    std::string propertyName = qualifiedCppName(
+            qualifiedQmlName(utils::deCapitalizeAsciiName(field->camelcase_name())));
     std::string propertyNameCap = utils::capitalizeAsciiName(propertyName);
 
     propertyMap["property_name"] = propertyName;
     propertyMap["property_name_cap"] = propertyNameCap;
     propertyMap["scriptable"] = scriptable;
+    propertyMap["export_macro"] = common::buildExportMacro();
 
     auto scopeTypeMap = produceMessageTypeMap(scope, nullptr);
     propertyMap["key_type"] = "";
     propertyMap["value_type"] = "";
-    propertyMap["classname"] = scope != nullptr ? scopeTypeMap["classname"] : "";
+    propertyMap["classname"] = scopeTypeMap["classname"];
+    propertyMap["dataclassname"] = propertyMap["classname"] + CommonTemplates::DataClassName();
     propertyMap["number"] = std::to_string(field->number());
 
+    if (common::isOneofField(field)) {
+        propertyMap["optional_property_name"] =
+                qualifiedCppName(qualifiedQmlName(field->containing_oneof()->name()));
+        propertyMap["optional_property_name_cap"] =
+                utils::capitalizeAsciiName(field->containing_oneof()->name());
+    } else if (common::isOptionalField(field)) {
+        propertyMap["optional_property_name"] = propertyName;
+        propertyMap["optional_property_name_cap"] = propertyNameCap;
+    }
+
     if (field->is_map()) {
         const Descriptor *type = field->message_type();
         auto keyMap = common::producePropertyMap(type->field(0), scope);
@@ -406,7 +595,13 @@ PropertyMap common::producePropertyMap(const FieldDescriptor *field, const Descr
     return propertyMap;
 }
 
-std::string common::qualifiedName(const std::string &name)
+std::string common::qualifiedCppName(std::string_view name)
+{
+    return utils::contains(CommonTemplates::ListOfCppExceptions(), name) ? std::string{ name } + "_"
+                                                                         : std::string{ name };
+}
+
+std::string common::qualifiedQmlName(std::string_view name)
 {
     std::string fieldName(name);
     const std::vector<std::string> &searchExceptions = CommonTemplates::ListOfQmlExceptions();
@@ -416,6 +611,34 @@ std::string common::qualifiedName(const std::string &name)
     return fieldName;
 }
 
+bool common::isOneofField(const FieldDescriptor *field)
+{
+#ifdef HAVE_PROTOBUF_SYNC_PIPER
+    return field->real_containing_oneof() != nullptr;
+#else
+    return field->containing_oneof() != nullptr;
+#endif
+}
+
+bool common::isTriviallyCopyable(const FieldDescriptor *field)
+{
+    return !field->is_repeated() &&
+        field->type() != FieldDescriptor::TYPE_MESSAGE &&
+        field->type() != FieldDescriptor::FieldDescriptor::TYPE_STRING &&
+        field->type() != FieldDescriptor::FieldDescriptor::TYPE_BYTES;
+}
+
+bool common::isOptionalField(const FieldDescriptor *field)
+{
+#ifdef HAVE_PROTOBUF_SYNC_PIPER
+    bool hasOptional = field->has_presence() && !field->real_containing_oneof();
+#else
+    bool hasOptional = field->file()->syntax() == FileDescriptor::SYNTAX_PROTO2
+        && field->is_optional() && !field->containing_oneof();
+#endif
+    return field->type() != FieldDescriptor::TYPE_MESSAGE && hasOptional;
+}
+
 bool common::isLocalEnum(const EnumDescriptor *type, const Descriptor *scope)
 {
     assert(type != nullptr);
@@ -446,6 +669,12 @@ bool common::hasQmlAlias(const FieldDescriptor *field)
             && Options::instance().hasQml();
 }
 
+bool common::hasNestedTypes(const Descriptor *type)
+{
+    return common::hasNestedMessages(type) || type->enum_type_count() > 0
+            || Options::instance().generateFieldEnum() || type->oneof_decl_count() > 0;
+}
+
 void common::iterateMessages(const FileDescriptor *file,
                              const std::function<void(const Descriptor *)> &callback)
 {
@@ -520,13 +749,14 @@ const Descriptor *common::findHighestMessage(const Descriptor *message)
 
 std::string common::collectFieldFlags(const FieldDescriptor *field)
 {
+    constexpr std::string_view flagsConstuctor = "uint(";
     std::string_view separator = " | ";
     std::string_view active_separator;
-    std::string flags;
+    std::string flags(flagsConstuctor);
 
     auto writeFlag = [&](const char *flag) {
         flags += active_separator;
-        flags += "QtProtobufPrivate::";
+        flags += "QtProtobufPrivate::FieldFlag::";
         flags += flag;
         active_separator = separator;
     };
@@ -534,13 +764,80 @@ std::string common::collectFieldFlags(const FieldDescriptor *field)
     if (field->type() != FieldDescriptor::TYPE_STRING
         && field->type() != FieldDescriptor::TYPE_BYTES
         && field->type() != FieldDescriptor::TYPE_MESSAGE
-        && field->type() != FieldDescriptor::TYPE_ENUM && !field->is_map() && field->is_repeated()
+        && !field->is_map() && field->is_repeated()
         && !field->is_packed()) {
         writeFlag("NonPacked");
     }
 
-    if (flags.empty())
-        writeFlag("NoFlags");
+    if (common::isOneofField(field))
+        writeFlag("Oneof");
 
+    if (common::isOptionalField(field))
+        writeFlag("Optional");
+
+    if (common::isOneofField(field) || common::isOptionalField(field)
+        || common::isPureMessage(field)) {
+        writeFlag("ExplicitPresence");
+    }
+
+    if (field->is_map())
+        writeFlag("Map");
+
+    if (field->is_repeated() && !field->is_map())
+        writeFlag("Repeated");
+
+    if (field->type() == FieldDescriptor::TYPE_ENUM)
+        writeFlag("Enum");
+
+    if (field->type() == FieldDescriptor::TYPE_MESSAGE)
+        writeFlag("Message");
+
+    if (flags == flagsConstuctor)
+        writeFlag("NoFlags");
+    flags += ")";
     return flags;
 }
+
+bool common::isExtraNamespacedFile(std::string_view file)
+{
+    return m_extraNamespacedFiles.find(std::string{ file }) != m_extraNamespacedFiles.end();
+}
+
+void common::setExtraNamespacedFiles(const std::set<std::string> &files)
+{
+    if (!files.empty() && files != m_extraNamespacedFiles)
+        m_extraNamespacedFiles = files;
+}
+
+std::string common::headerGuardFromFilename(std::string fileName)
+{
+    utils::asciiToUpper(fileName);
+    return utils::replace(fileName, ".", "_");
+}
+
+std::string common::generateRelativeFilePath(const FileDescriptor *file, const std::string &name)
+{
+    std::string outFileBasename;
+    if (Options::instance().isFolder()) {
+        outFileBasename = file->package();
+        if (!outFileBasename.empty()) {
+            outFileBasename = utils::replace(outFileBasename, ".", "/");
+            outFileBasename += '/';
+        }
+    }
+    outFileBasename += name;
+
+    return outFileBasename;
+}
+
+bool common::hasCustomJsonCoversion(const Descriptor *message)
+{
+    static constexpr std::array<std::string_view, 11> TypesSupportingOptionalCoversion{
+        "Timestamp",   "Duration",   "BoolValue",   "Int32Value", "Int64Value",  "UInt32Value",
+        "UInt64Value", "FloatValue", "DoubleValue", "BytesValue", "StringValue",
+    };
+
+    return std::find(TypesSupportingOptionalCoversion.begin(),
+                     TypesSupportingOptionalCoversion.end(), std::string_view(message->name()))
+        != TypesSupportingOptionalCoversion.end();
+}
